#include "CompBioTools.hh"
#include <string>
#include "zipiostream.hh"


#include "AceFile.hh"


string gVersion = "1.02";

//---------------------------------------------------
// * PrintVersion
//---------------------------------------------------
//
void PrintVersion()
{
  cout<<"
    Version "<<gVersion<<endl;
}

string OptAceFileName="";
string OptContigQualityFileName="";
string OptReadQualityFileName="";
string OptRootFileName="";

int OptBasesPerLine = 100;
bool OptShowQuality = false;

int OptPhredTrim = 20;
int OptContigTrim = 1;
int OptStart = 0;
int OptEnd = -1;
int OptContigNum = -1;

bool  OptStatisticsOnly = false;
bool  OptTemplateInfo = false;

//---------------------------------------------------
// * PrintHelp
//---------------------------------------------------
//
void PrintHelp()
{
printf("\n

    bcm-ace-explore 

    Prints the multiple alignment for the specified contigs in an ace file
    Useful if you don't have phd files or x-windows or other things that consed 
    needs.  Requires just these two files as input:

         * Ace File  (e.g. phrap.seq.ace)
         * Read Quality File (e.g. phrap.seq.qual)

    Files can either be plain or gzipped (gzipped files must end in .gz). 


    -a  AceFileName
    -r  ReadQualityFileName {optional, needed for some stats and quality print out). 

    -n  ContigNum       Which contig to print (default will print all of them). 
    -s  StartBase       Index of first base within contig to print. 
    -e  EndBase         Index of last base within contig to print. (default is all bases). 

    -P  PhredCutoff     Phred value for computing read trims.  (default = 20);
    -C  ContigCutoff    Phrap value for computing contig trims (default = 1);


    -Q                  Show quality in output. (defalt = don't show). This option 
                        will take 3 times as much space per base, so adjust BasesPerLine accordingly. 

    -l  BasesPerLine    Bases to show per line of output (default = 100).

    -A   AceRoot        Overrides -a, and -r.  Attempts to derive the ace, contig quality, 
                        and read quality file names from a root name. Assumes ace file is AceRoot.ace, 
                        that contig quality file is AceRoot.contigs.qual, and that reads 
                        quality file is AceRoot.  Example, AceRoot=consed.fasta, then 
                        ace file is assumed to be consed.fasta.ace, contig quality file 
                        is assumed to be consed.fasta.contigs.qual, and read quality file is 
                        assumed to be consed.fasta.qual.   It checks for both the ungzipped 
                        version and the gzipped version (which must end in .gz).  Assumes that
                        if ace file is zipped, all three files are zipped. 

      ----------------------------------------

     -S                 Statistics only.  Has no effect other than to turn off printing of 
                        contig alignments. 

     -T                 Print template info.  



     Key to symbols in print out:

         . Read does not extend to the given position. 
         - Quality trimmed region. 
         % X trimmed region (usually X also low quality so rarely see this.) 

    Written by: James Durbin (kdurbin@bcm.tmc.edu)
");

 PrintVersion();

 printf(" 
    Version History:  
    ----------------
    1.02   Changed to get consensus quality from .ace file rather than from contigs.qual file. 
    1.00   Initial release. 
");
}


//---------------------------------------------------
// * SetupOptions
//---------------------------------------------------
//
void SetupOptions(int argc, char**argv)
{
  // Create parser object and define valid options
  // Options with a : require a parameter. 
  GetOption Parser(argc,argv,"a:r:Ql:A:hC:P:s:e:n:TS");

  // Handle the options...
  char theOption;
  while((theOption = Parser.NextOption()) != OPTION_EOF){
    switch(theOption){
	 case 'a': OptAceFileName = Parser.GetStringArg();break;
	 case 'r': OptReadQualityFileName = Parser.GetStringArg();break;
	 case 'Q': OptShowQuality = true;break;
	 case 'l': OptBasesPerLine = Parser.GetIntArg();break;
	 case 'A': OptRootFileName = Parser.GetStringArg();break;
	 case 'P': OptPhredTrim = Parser.GetIntArg();break;
	 case 'C': OptContigTrim = Parser.GetIntArg();break;
	 case 's': OptStart = Parser.GetIntArg();break;
	 case 'e': OptEnd = Parser.GetIntArg();break;
	 case 'n': OptContigNum = Parser.GetIntArg();break;
	 case 'S': OptStatisticsOnly = true;break;
	 case 'T': OptTemplateInfo = true;break;
	 default: PrintHelp();exit(1);break;
    }
  }
}

//bool lt_templateInfo(TemplateInfo t1,TemplateInfo t2){
//  return(t1.mFwdContig<t2.mFwdContig);
//};

//-------------------------------------------
// * main
//-------------------------------------------
//
int main(int argc,char **argv)
{
  AceFile Ace;
  FASTAFile FA;

  SetupOptions(argc,argv);

  Ace.mPhredTrim = OptPhredTrim;
  Ace.mContigTrim = OptContigTrim;
  Ace.mMinContigLength = 800;
  
  if (OptRootFileName != ""){
	 Ace.ReadAll(OptRootFileName);
  }else{
	 if (OptReadQualityFileName == ""){
		Ace.mPhredTrim = 0;
		Ace.Read(OptAceFileName);
	 }else{
		Ace.ReadAll(OptAceFileName,OptReadQualityFileName);
	 }
  }
  cout<<endl;

  if (OptContigNum >= 0){
	 cout<<"Ace["<<OptContigNum<<"].size="<<Ace[OptContigNum].size()<<endl;
  }

  if (!OptStatisticsOnly){
	 if (OptShowQuality){
		if (OptContigNum < 0){
		  for(int c = 0;c< Ace.size();c++){
			 //cout<<"Contig"<<c<<endl;
			 Ace[c].mPrintLineSize = OptBasesPerLine;
			 Ace[c].PrintAlignmentWithQuality(cout,OptStart,OptEnd);
			 cout<<endl;
		  }
		}else{
		  if (OptContigNum < Ace.size()){
			 Ace[OptContigNum].mPrintLineSize = OptBasesPerLine;
			 Ace[OptContigNum].PrintAlignmentWithQuality(cout,OptStart,OptEnd);
		  }else{
			 cout<<"Ace file only has "<<Ace.size()<<" contigs."<<endl;		  
		  }
		}

	 }else{
		if (OptContigNum < 0){
		  for(int c = 0;c< Ace.size();c++){
			 //cout<<"Contig"<<c<<endl;
			 Ace[c].mPrintLineSize = OptBasesPerLine;
			 Ace[c].PrintAlignment(cout,OptStart,OptEnd);
			 cout<<endl;
		  }
		}else{
		  if (OptContigNum < Ace.size()){
			 Ace[OptContigNum].mPrintLineSize = OptBasesPerLine;
			 Ace[OptContigNum].PrintAlignment(cout,OptStart,OptEnd);
		  }else{
			 cout<<"Ace file only has "<<Ace.size()<<" contigs."<<endl;		  
		  }
		}
	 }
  }else{
	 if (OptContigNum < Ace.size()){
		Ace[OptContigNum].PrintTemplateInfo(cout);
	 }else{
		cout<<"Ace file only has "<<Ace.size()<<" contigs."<<endl;		  
	 }
  }
}
