#include "CBT.hh"
#include <cstdlib>
#include <cmath>

class summary
{

private:
    double count, sum_err, sum_pred_err;

public:

    summary() : count(0), sum_err(0.0), sum_pred_err(0.0) {};

    void add(int err) 
    {
        count++; 
        sum_err+=err;
    }

    void add(int err, double base_qual) 
    {
        add(err); 
        sum_pred_err += pow(10.0, (base_qual/-10.0)); //sum predicted_errors
    }



    friend ostream & operator << (ostream &o, const summary &s);
};

ostream & operator << (ostream &o, const summary &s)
{
    o << s.count << ", " << s.sum_err;

    if (s.sum_pred_err > 0){
        o << ", " << s.sum_pred_err;
    }


    double sum_err = s.sum_err;

    if (sum_err < 1){
        // to avoid missing data, assume no read is perfect
        sum_err = 1;
    }

    // average error as phred score

    cout << ", " << -10 * log10(sum_err/s.count); 

    // average predicted error shown as phred score

    if (s.sum_pred_err > 0)
        cout << ", " << -10*log10(s.sum_pred_err/s.count); 

    return o;
};

template<class T> void print_summary(const T &summap)
{
    typename T::const_iterator it;
    for (it = summap.begin(); it != summap.end(); ++it) {
        cout << it->first << ", " << it->second << "\n";
    }   
    cout << "\n";
}
 

// IMPORTANT NOTE:  
//
// I use ZERO BASED COORDINATES everywhere in my code always.
//  
// Ace files, crossmach, phrap, and the whole biology world uses ONE BASED COORDINATES. 
// I always use ZERO BASED internally, and convert from/to one based when I read in 
// or write out something. 
// 
// 

//----------------------------------------
// * GetBaseAndQualityAtConsensus
//----------------------------------------
// 
// Gets the bases and qualities for the given consensus position. 
// 
// 
void GetBaseAndQuality(Contig &C, int ConsensusBaseIdx, vector<char> &bases, vector<int> &qualities, vector<int> &positions,
		       vector<string> &names){
  
  for(int ReadNum = 0;ReadNum < C.mReads.size();ReadNum++){    
    
    // Figure out what the read coordinate for this consensus position is.
    int ReadBaseIdx = C.CIdx2RIdx(ConsensusBaseIdx,ReadNum);

    // Are these coordinates actually inside the read?  If so add the base and quality. 

    // Inside read is defined relative to the left and right trim points or you can define 
    // it with respect to the start and end of the read.  

    //if ((readIdx >= mReadTrimLeft[readNum]) && (readIdx <= mReadTrimRight[readNum])){

    if ((ReadBaseIdx >= 0) && (ReadBaseIdx < C.mReads[ReadNum].size())){
      bases.push_back(C.mReads[ReadNum][ReadBaseIdx]);
      qualities.push_back(C.mReadsQuality[ReadNum][ReadBaseIdx]);
      positions.push_back(ReadBaseIdx);
      names.push_back(C.mReads[ReadNum].mName);
    }
  }
}


//----------------------------------------
// * process
//----------------------------------------
// 
// Process the given .ace and given .qual file
// to extract summary statistics on matching bases 
// 
// 
void process(const string &AceIn, const string &QualIn)
{ 
  AceFile Ace;  

  // ReadAll automatically runs both contig and read trims.  These are crude 
  // trims that just scan until they find the first base with >= the min quality. 
  // Seting min quality to 0 effectively turns off the trim.  For your purposes
  // you will probably want to turn it off. 
  Ace.mPhredTrim = 0;  
  Ace.mContigTrim = 0;

  Ace.ReadAll(const_cast<string &>(AceIn), const_cast<string &>(QualIn));

  map<int, summary> by_posn;
  map<int, summary> by_base_qual;
  map<char, summary> by_base;

  // For each contig in the ace file...
  for(int i = 0;i < Ace.size();i++){
	// cout << "Contig " << i+1 << endl;

    Contig &C = Ace[i];

    // For each consensus position in that contig...
    for(int cIdx = 0;cIdx < C.size();cIdx++){
      vector<char> bases;
      vector<int> qualities;
      vector<int> positions;
      vector<string> rnames;
      GetBaseAndQuality(C,cIdx,bases,qualities, positions,rnames);
      for(int b = 0;b < bases.size();b++){
        char consed = toupper(C[cIdx]);
        char base = toupper(bases[b]);
        int base_qual = qualities[b];
        // Get consensus quality
        int phrapqual = C.mConsensusQuality[cIdx];

        if (phrapqual < 80) // cutoff for assuming that consensus is correct
          continue;

        if (base == 'N' || base == '*' || base == 'X' ||  base_qual < 5)
          continue;

        if (consed == 'N' || consed == '*' || consed == 'X' )
          continue;

        if (!isalpha(base)){
	  cerr << "invalid base code of '" << base << "' at "<<C.mName<< " Contig position: " <<cIdx+1<<"\treadName: "<<rnames[b]<<" pos: "<<positions[b]<<endl;
            continue;
        }

        if (base == 'Z')
        {            
	  cerr << "invalid base code of Z at Contig #:" << i+1 << " Contig position: " << cIdx+1 << "\n";
	  continue;
        }

        if (base_qual > 70){
            cerr << "invalid quality of " << base_qual << " at Contig #:" << i+1 << " Contig position: " << cIdx+1 << "\n";
            continue;
        }

        int err;
        if (consed != base){
          // cout << cIdx+1 << "," << C[cIdx] << "," << phrapqual << "," << bases[b] << "," << qualities[b] << "\n";
	  err = 1;
	}
        else{
	  err = 0;
	}

        // summarize
        //
        // by position
        by_posn[positions[b]].add(err, base_qual);
        
        // by base quality
        by_base_qual[base_qual].add(err);
      
        // by base
        by_base[base].add(err);

      }
    }
  }

  // Write out all of those beautiful stats...
  cout << "base_qual, count, sum_errors, actual_qual\n";
  print_summary(by_base_qual);

  cout << "posn, count, sum_errors, sum_predicted_errors, actual_qual, predicted_qual\n";
  print_summary(by_posn);

  cout << "base_read, count, sum_errors, actual_qual\n";
  print_summary(by_base);
}


int main(int argc, char** argv){

  //#if 0
  if (argc < 2){
    cout<<"cjoytest  AceIn ReadsQuality\n"<<endl;
    exit(1);
  }

  string AceIn = argv[1];
  string QualIn = argv[2];
  //#endif

  //string base_dir = "/cygdrive/d/mouse_testing/mbbh_phred/edit_dir/";
  // string base_dir = "/data/mouse/testing/mbbh.phred/edit_dir/";

  //string AceIn = base_dir + "mbbh.phred.fasta.screen.ace.1";
  //string QualIn = base_dir + "mbbh.phred.fasta.screen.qual";

  
  process(AceIn, QualIn);
  return 0;

}

















